#!/usr/bin/perl

if ($#ARGV < 0) {
  print "\n\nUsage:\n\namber2gmx.pl parmXX.dat\n\n";
  print "The argument is the name of the amber force field parameter file to use\n";
  print "for the conversion. The output files will be called ffamber - rename it\n";
  print "to reflect the version of the Amber forcefield you converted.\n";
  exit;
}

open(AMBERPARM,$ARGV[0]) or die "Couldn't open the Amber parameter file '$ARGV[0]': $!\n";
open(ATP,">ffamber.atp") or die "Couldn't open output file: $!\n";
open(NB,">ffambernb.itp") or die "Couldn't open output file: $!\n";
open(BONDED,">ffamberbon.itp") or die "Couldn't open output file: $!\n";

# The first line is a comment/description
$ffcomment=<AMBERPARM>;

print "Processing atomtypes...\n";
# Then we have all the atom types
while(($line=<AMBERPARM>) && ($line =~ /^[A-Z]/)) {
  @data=split(" ",$line);
  $thistype=@data[0];
  $thismass=@data[1];
  $comment=substr $line, 37;
  chomp($comment);
  push @atomtype,$thistype;
  push @atommass,$thismass;
  printf ATP " %-8s %8.2f  ; %s\n",$thistype,$thismass,$comment;
}

<AMBERPARM>; # skip a line with atomtypes

print BONDED "[ bondtypes ]\n;    i     j    func       b0          kb\n";

print "Processing bondtypes...\n";
# read bonds
while(($line=<AMBERPARM>) && ($line =~ /^[A-Z]/)) {
  @data=split(" ",substr($line,6)); # skip the atom types at the beginning
  $atom1=substr($line,0,2);
  $atom2=substr($line,3,2);
  $forceconst=$data[0]*2*4.1868*100; # convert kcal/aa^2 -> kJ/nm, and a factor 2 in the definition
  $bondlength=0.1*$data[1]; # convert AA->nm
  $comment=substr($line,28);
  chomp($comment);
  printf BONDED "  %4s  %4s     1  %9.5f    %8.1f ; %s\n",$atom1,$atom2,$bondlength,$forceconst,$comment;
}

print "Processing angletypes...\n";
print BONDED "\n\n\n[ angletypes ]\n;    i     j     k  func       th0        cth\n";
# read angles
while(($line=<AMBERPARM>) && ($line =~ /^[A-Z]/)) {
  @data=split(" ",substr($line,10)); # skip the atom types at the beginning
  $atom1=substr($line,0,2);
  $atom2=substr($line,3,2);
  $atom3=substr($line,6,2);
  $forceconst=$data[0]*2*4.1868; # kcal->kJ, and a factor 2 in the definition
  $angle=$data[1]; 
  $comment=substr($line,32);
  chomp($comment);
  printf BONDED "  %4s  %4s  %4s    1   %8.2f   %8.1f ; %s\n",$atom1,$atom2,$atom3,$angle,$forceconst,$comment;
}

print "Processing dihedraltypes...\n";
print BONDED "\n\n\n[ dihedraltypes ]\n;   i     j     k     l   func     coefficients\n";
print BONDED "; Some dihedrals need the R-B format, so we use it for all...\n";
# read angles
$cont=0;
$V[0]=$V[1]=$V[2]=$V[3]=$V[4]=$V[5]=0;
$phi[0]=$phi[1]=$phi[2]=$phi[3]=$phi[4]=$phi[5]=0;

while(($line=<AMBERPARM>) && ($line =~ /^[A-Z]/)) {
  @data=split(" ",substr($line,12)); # skip the atom types at the beginning
  if($cont) {
      $a1=substr($line,0,2);
      $a2=substr($line,3,2);
      $a3=substr($line,6,2);
      $a4=substr($line,9,2);
      if(($a1 ne $atom1) || ($a2 ne $atom2) || ($a3 ne $atom3) || ($a4 ne $atom4)) {
	  print "Error - nonmatching atoms for dihedrals\n";
	  exit;
      }
      $numdih=$data[0];
      $forceconst=$data[1];
      $thisphi=$data[2];
      $mult=$data[3];
      if($mult<0) {
	  $cont=1;
	  $mult=-$mult;
      } else {
	  $cont=0;
      }
      $V[$mult]=$forceconst*4.1868;      
      if((($mult==1) || ($mult==3)) && $thisphi==180) {
	  $thisphi=0;
	  $V[$mult]=-$V[$mult];
      } 
      if((($mult==2) || ($mult==4)) && $thisphi==0) {
	  $thisphi=180;
	  $V[$mult]=-$V[$mult];
      } 
      if (((($mult==2) || ($mult==4)) && ($thisphi!=180)) ||
	  ((($mult==1) || ($mult==3)) && ($thisphi!=0))) {
	  print "Error - dihedral phase must be 0 for odd and 180 for even term.\n";
	  exit;
      }    
      if(!$cont) {
	  $C0=sprintf "%8.3f",$V[2]+0.5*($V[1]+$V[3]);
	  $C1=sprintf "%8.3f",0.5*(3.0*$V[3]-$V[1]);
	  $C2=sprintf "%8.3f",4.0*$V[4]-$V[2];
	  $C3=sprintf "%8.3f",0.0-2.0*$V[3];
	  $C4=sprintf "%8.3f",0.0-4.0*$V[4];
	  $C5=sprintf "%8.3f",0.0;
	  printf BONDED "  %4s  %4s  %4s  %4s    3    $C0 $C1 $C2 $C3 $C4 $C5 ; %s\n",
	  $atom1,$atom2,$atom3,$atom4,$comment;
	  $cont=0;
	  $V[0]=$V[1]=$V[2]=$V[3]=$V[4]=$V[5]=0;
	  $phi[0]=$phi[1]=$phi[2]=$phi[3]=$phi[4]=$phi[5]=0;
      }
  } else {
      $atom1=substr($line,0,2);
      $atom2=substr($line,3,2);
      $atom3=substr($line,6,2);
      $atom4=substr($line,9,2);
      $numdih=$data[0];
      $forceconst=$data[1];
      $thisphi=$data[2];
      $mult=$data[3];
      $comment=substr($line,60);
      chomp($comment);
      if($mult<0) {
	  $cont=1;
	  $mult=-$mult;
      }
      $V[$mult]=$forceconst*4.1868;      
      if((($mult==1) || ($mult==3)) && $thisphi==180) {
	  $thisphi=0;
	  $V[$mult]=-$V[$mult];
      } 
      if((($mult==2) || ($mult==4)) && $thisphi==0) {
	  $thisphi=180;
	  $V[$mult]=-$V[$mult];
      } 
      if (((($mult==2) || ($mult==4)) && ($thisphi!=180)) ||
	  ((($mult==1) || ($mult==3)) && ($thisphi!=0))) {
	  print "Error - dihedral phase must be 0 for odd and 180 for even term.\n";
	  exit;
      }    
      if(!$cont) {
	  $C0=sprintf "%8.3f",$V[2]+0.5*($V[1]+$V[3]);
	  $C1=sprintf "%8.3f",0.5*(3.0*$V[3]-$V[1]);
	  $C2=sprintf "%8.3f",4.0*$V[4]-$V[2];
	  $C3=sprintf "%8.3f",0.0-2.0*$V[3];
	  $C4=sprintf "%8.3f",0.0-4.0*$V[4];
	  $C5=sprintf "%8.3f",0.0;
	  printf BONDED "  %4s  %4s  %4s  %4s    3    $C0 $C1 $C2 $C3 $C4 $C5 ; %s\n",
	  $atom1,$atom2,$atom3,$atom4,$comment;
	  $cont=0;
	  $V[0]=$V[1]=$V[2]=$V[3]=$V[4]=$V[5]=0;
	  $phi[0]=$phi[1]=$phi[2]=$phi[3]=$phi[4]=$phi[5]=0;
      }
  }
}


print "Processing impropers...\n";
print BONDED "\n\n[ dihedraltypes ]\n;   i     j     k     l   func     coefficients\n";
print BONDED "; Amber impropers are periodic functions - represented with Gromacs proper dihedrals\n";

# read angles
while(($line=<AMBERPARM>) && ($line =~ /^[A-Z]/)) {
  @data=split(" ",substr($line,14)); # skip the atom types at the beginning
  $atom1=substr($line,0,2);
  $atom2=substr($line,3,2);
  $atom3=substr($line,6,2);
  $atom4=substr($line,9,2);
  $forceconst=$data[0]*0.5*4.1868; # Amber uses 1-cos(2x), while Gmx proper is [1+cos(2*x+180)] (Magnitude differs!)
  $comment=substr($line,60);
  chomp($comment);
  printf BONDED "  %4s  %4s  %4s  %4s    1      180.0 %8.1f    2 ; %s\n",$atom1,$atom2,$atom3,$atom4,$forceconst,$comment;
}

while(($line=<AMBERPARM>) && !($line =~ /^[A-Z]/)) {}

# Get two lines of Lennard-Jones equivalencies
@Nequiv=split(" ",$line);
@Cequiv=split(" ",<AMBERPARM>);

while(($line=<AMBERPARM>) && !($line =~ /^[A-Z]/)) {}

print "Processing nonbonded parameters...\n";
print NB "\n[ atomtypes ]\n; name       mass    charge   ptype          sigma       epsilon\n";
# read angles

while(($line=<AMBERPARM>) && ($line =~ /^  [A-Z]/)) {
  @data=split(" ",$line);
  $atom=$data[0];
  $rmin=$data[1];
  $eps=$data[2]; 
  $comment=substr($line,40);
  chomp($comment);
  push @nbatomtypes, $atom;
  $sig=0.1*(2.0*$rmin/(2**(1.0/6.0)));
  $eps=$eps*4.1868;
  push @sigma, $sig;
  push @epsilon, $eps;
  push @nbcomments, $comment;  
}

# Check which atom in the equivalency array has L-J parameters
for($i=0;$i<=$#Nequiv && !$found;$i++) {
  $found=0;
  for($j=0;$j<=$#nbatomtypes && !$found;$j++) {
    $found=($nbatomtypes[$j] eq $Nequiv[$i]);
  }
}
if(!$found) {
  print "No atom type with L-J parameters found in equivalency list.";
  exit;
}
$Nequivmain=$j;

for($i=0;$i<=$#Cequiv && !$found;$i++) {
  $found=0;
  for($j=0;$j<=$#nbatomtypes && !$found;$j++) {
    $found=($nbatomtypes[$j] eq $Nequiv[$i]);
  }
}
if(!$found) {
  print "No atom type with L-J parameters found in equivalency list.";
  exit;
}
$Cequivmain=$j;

# Go through the atomtype array and write out sig/eps for each type.
for($i=0;$i<=$#atomtype;$i++) {
  # first check if we are in the equiv list.
  $found=0;
  for($j=0;$j<=$#Cequiv && !$found;$j++) {
    $found=($Cequiv[$j] eq $atomtype[$i]);
  }
  if($found) {
    $num=$Cequivmain;
  } else {
    for($j=0;$j<=$#Nequiv && !$found;$j++) {
      $found=($Nequiv[$j] eq $atomtype[$i]);
    }
    if($found) {
      $num=$CNequivmain;
    }
  }
  if(!$found) {
    for($j=0;$j<$#nbatomtypes && !$found;$j++) {
      $found=($atomtype[$i] eq $nbatomtypes[$j]);
      $num=$j;
    }
  }
  if(!$found) {
    print "Warning: No L-J param for atomtype $atomtype[$i]\n";
    printf NB "  %4s  %9.2f     0.000      A     %12.5e  %12.5e ; Unknown parameters!\n",$atomtype[$i],$atommass[$i],0,0;    
  } else {
    printf NB "  %4s  %9.2f     0.000      A     %12.5e  %12.5e\n",$atomtype[$i],$atommass[$i],$sigma[$num],$epsilon[$num];
 }
}



print "\n\nDone. You will have to set up the following things manually:\n\n";
print "* Rename the forcefield files\n";
print "* Setup the [defaults] section in the main ffamberXX.itp include the bon/nb files\n";
print "* Heavy hydrogen masses\n";
print "* Dummy constraints\n";
print "* L-J parameters for any atoms that missed then (in case you got a warning above)\n";
print "* Sort dihedrals so that wildcarded types come after non-wildcarded ones.\n";
